# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""IntOption
"""

from ..i18n import _
from .option import Option


class IntOption(Option):
    "represents a choice of an integer"
    __slots__ = tuple()
    _type = _('integer')

    def __init__(self,
                 *args,
                 min_number=None,
                 max_number=None,
                 **kwargs):
        extra = {}
        if min_number is not None:
            extra['min_number'] = min_number
        if max_number is not None:
            extra['max_number'] = max_number
        super().__init__(*args, extra=extra, **kwargs)

    def validate(self,
                 value: int,
                 ) -> None:
        if not isinstance(value, int):
            raise ValueError()

    def second_level_validation(self,
                                value,
                                warnings_only):
        min_number = self.impl_get_extra('min_number')
        if min_number is not None and value < min_number:
            if warnings_only:
                msg = 'value should be greater than "{0}"'
            else:
                msg = 'value must be greater than "{0}"'
            raise ValueError(_(msg).format(min_number))
        max_number = self.impl_get_extra('max_number')
        if max_number is not None and value > max_number:
            if warnings_only:
                msg = 'value should be less than "{0}"'
            else:
                msg = 'value must be less than "{0}"'
            raise ValueError(_(msg).format(max_number))



# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough gus of pypy: pypy: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"enables us to carry out a calculation and return an option's value"
from typing import Any, Optional, Union, Callable, Dict, List
from itertools import chain

from .error import PropertiesOptionError, ConfigError, LeadershipError, ValueWarning
from .i18n import _
from .setting import undefined, ConfigBag, OptionBag, Undefined
from .function import FUNCTION_WAITING_FOR_DICT
# ____________________________________________________________


class Params:
    __slots__ = ('args', 'kwargs')
    def __init__(self, args=None, kwargs=None, **kwgs):
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = {}
        if kwgs:
            kwargs.update(kwgs)
        if isinstance(args, Param):
            args = (args,)
        else:
            if not isinstance(args, tuple):
                raise ValueError(_('args in params must be a tuple'))
            for arg in args:
                if not isinstance(arg, Param):
                    raise ValueError(_('arg in params must be a Param'))
        if not isinstance(kwargs, dict):
            raise ValueError(_('kwargs in params must be a dict'))
        for arg in kwargs.values():
            if not isinstance(arg, Param):
                raise ValueError(_('arg in params must be a Param'))
        self.args = args
        self.kwargs = kwargs


class Param:
    pass


class ParamOption(Param):
    __slots__ = ('option',
                 'notraisepropertyerror',
                 'raisepropertyerror',
                 )
    def __init__(self,
                 option: 'Option',
                 notraisepropertyerror: bool=False,
                 raisepropertyerror: bool=False,
                 ) -> None:
        if __debug__ and not hasattr(option, 'impl_is_symlinkoption'):
            raise ValueError(_('paramoption needs an option not {}').format(type(option)))
        if option.impl_is_symlinkoption():
            cur_opt = option.impl_getopt()
        else:
            cur_opt = option
        assert isinstance(notraisepropertyerror, bool), _('param must have a boolean not a {} for notraisepropertyerror').format(type(notraisepropertyerror))
        assert isinstance(raisepropertyerror, bool), _('param must have a boolean not a {} for raisepropertyerror').format(type(raisepropertyerror))
        self.option = cur_opt
        self.notraisepropertyerror = notraisepropertyerror
        self.raisepropertyerror = raisepropertyerror


class ParamDynOption(ParamOption):
    __slots__ = ('subpath',)
    def __init__(self,
                 option: 'Option',
                 subpath: str,
                 dynoptiondescription: 'DynOptionDescription',
                 notraisepropertyerror: bool=False,
                 raisepropertyerror: bool=False,
                 optional: bool=False,
                 ) -> None:
        super().__init__(option,
                         notraisepropertyerror,
                         raisepropertyerror,
                         )
        self.subpath = subpath
        self.dynoptiondescription = dynoptiondescription
        self.optional = optional


class ParamSelfOption(Param):
    __slots__ = ('whole')
    def __init__(self,
                 whole: bool=undefined,
                 ) -> None:
        """whole: send all value for a multi, not only indexed value"""
        if whole is not undefined:
            self.whole = whole


class ParamValue(Param):
    __slots__ = ('value',)
    def __init__(self, value):
        self.value = value


class ParamInformation(Param):
    __slots__ = ('information_name',
                 'default_value',
                 'option',
                 )
    def __init__(self,
                 information_name: str,
                 default_value: Any=undefined,
                 option: 'Option'=None
                 ) -> None:
        self.information_name = information_name
        self.default_value = default_value
        if option:
            if option.impl_is_symlinkoption():
                raise ValueError(_('option in ParamInformation cannot be a symlinkoption'))
            if option.impl_is_follower():
                raise ValueError(_('option in ParamInformation cannot be a follower'))
            if option.impl_is_dynsymlinkoption():
                raise ValueError(_('option in ParamInformation cannot be a dynamic option'))
        self.option = option


class ParamSelfInformation(ParamInformation):
    __slots__ = tuple()


class ParamIndex(Param):
    __slots__ = tuple()


class ParamSuffix(Param):
    __slots__ = tuple()


class Calculation:
    __slots__ = ('function',
                 'params',
                 'help_function',
                 '_has_index',
                 'warnings_only',
                 )
    def __init__(self,
                 function: Callable,
                 params: Params=Params(),
                 help_function: Optional[Callable]=None,
                 warnings_only: bool=False,
                 ):
        assert isinstance(function, Callable), _('first argument ({0}) must be a function').format(function)
        if help_function:
            assert isinstance(help_function, Callable), _('help_function ({0}) must be a function').format(help_function)
            self.help_function = help_function
        else:
            self.help_function = None
        self.function = function
        self.params = params
        for arg in chain(self.params.args, self.params.kwargs.values()):
            if isinstance(arg, ParamIndex):
                self._has_index = True
                break
        if warnings_only is True:
            self.warnings_only = warnings_only

    def execute(self,
                option_bag: OptionBag,
                orig_value: Any=undefined,
                allow_value_error: bool=False,
                force_value_warning: bool=False,
                for_settings: bool=False,
                ) -> Any:
        return carry_out_calculation(option_bag.option,
                                     callback=self.function,
                                     callback_params=self.params,
                                     index=option_bag.index,
                                     config_bag=option_bag.config_bag,
                                     orig_value=orig_value,
                                     allow_value_error=allow_value_error,
                                     force_value_warning=force_value_warning,
                                     for_settings=for_settings,
                                     )

    def help(self,
             option_bag: OptionBag,
             for_settings: bool=False,
             ) -> str:
        if not self.help_function:
            return self.execute(option_bag,
                                for_settings=for_settings,
                                )
        return carry_out_calculation(option_bag.option,
                                     callback=self.help_function,
                                     callback_params=self.params,
                                     index=option_bag.index,
                                     config_bag=option_bag.config_bag,
                                     for_settings=for_settings,
                                     )

    def __deepcopy__(x, memo):
        return x


def manager_callback(callback: Callable,
                     param: Param,
                     option,
                     index: Optional[int],
                     orig_value,
                     config_bag: ConfigBag,
                     for_settings: bool,
                     ) -> Any:
    """replace Param by true value"""
    def calc_index(param, index, same_leadership):
        if index is not None:
            if hasattr(param, 'whole'):
                whole = param.whole
            else:
                # if value is same_leadership, follower are isolate by default
                # otherwise option is a whole option
                whole = not same_leadership
            if not whole:
                return index
        return None

    def calc_self(param,
                  option,
                  index,
                  value,
                  config_bag,
                  ):
        # index must be apply only if follower
        is_follower = option.impl_is_follower()
        apply_index = calc_index(param, index, is_follower)
        if value is undefined or (apply_index is None and is_follower):
            path = option.impl_getpath()
            option_bag = OptionBag(option,
                                   None,
                                   config_bag,
                                   properties=None,
                                   )
            properties = config_bag.context.get_settings().getproperties(option_bag,
                                                                         uncalculated=True,
                                                                         )
            new_value = get_value(config_bag,
                                  option,
                                  param,
                                  apply_index,
                                  True,
                                  properties,
                                  )
            if apply_index is None and is_follower:
                new_value[index] = value
            value = new_value
        elif apply_index is not None and not is_follower:
            value = value[apply_index]
        return value

    def get_value(config_bag,
                  option,
                  param,
                  index,
                  self_calc,
                  properties=undefined,
                  ):
        parent_option_bag, option_bag = get_option_bag(config_bag,
                                                       option,
                                                       param,
                                                       index,
                                                       self_calc,
                                                       properties=properties,
                                                       )
        if option.impl_is_follower() and index is None:
            value = []
            for idx in range(config_bag.context.get_length_leadership(parent_option_bag)):
                parent_option_bag, option_bag = get_option_bag(config_bag,
                                                               option,
                                                               param,
                                                               idx,
                                                               self_calc,
                                                               properties=properties,
                                                               )
                value.append(_get_value(param,
                                        option_bag,
                                        ))
        else:
            value = _get_value(param,
                               option_bag,
                               )
        return value

    def _get_value(param: Params,
                   option_bag: OptionBag,
                   ) -> Any:
        try:
            # get value
            value = config_bag.context.get_value(option_bag)
        except PropertiesOptionError as err:
            # raise PropertiesOptionError (which is catched) because must not add value None in carry_out_calculation
            if param.notraisepropertyerror or param.raisepropertyerror:
                raise err from err
            raise ConfigError(_('unable to carry out a calculation for "{}"'
                                ', {}').format(option.impl_get_display_name(), err), err) from err
        except ValueError as err:
            raise ValueError(_('the option "{0}" is used in a calculation but is invalid ({1})').format(option_bag.option.impl_get_display_name(), err)) from err
        except AttributeError as err:
            if isinstance(param, ParamDynOption) and param.optional:
                # cannot acces, simulate a propertyerror
                raise PropertiesOptionError(option_bag,
                                            ['configerror'],
                                            config_bag.context.get_settings(),
                                            )
            raise ConfigError(_(f'unable to get value for calculating "{option_bag.option.impl_get_display_name()}", {err}')) from err
        return value

    def get_option_bag(config_bag,
                       opt,
                       param,
                       index_,
                       self_calc,
                       properties=undefined,
                       ):
        # don't validate if option is option that we tried to validate
        config_bag = config_bag.copy()
        if for_settings:
            config_bag.properties = config_bag.true_properties - {'warnings'}
        config_bag.set_permissive()
        if not for_settings:
            config_bag.properties -= {'warnings'}
        if self_calc:
            config_bag.unrestraint()
            config_bag.remove_validation()
        root_option_bag = OptionBag(config_bag.context.get_description(),
                                    None,
                                    config_bag,
                                    )
        try:
            options_bag = config_bag.context.get_sub_option_bag(root_option_bag,
                                                                opt.impl_getpath(),
                                                                index_,
                                                                validate_properties=not self_calc,
                                                                properties=properties,
                                                                )
        except PropertiesOptionError as err:
            # raise PropertiesOptionError (which is catched) because must not add value None in carry_out_calculation
            if param.notraisepropertyerror or param.raisepropertyerror:
                raise err from err
            raise ConfigError(_('unable to carry out a calculation for "{}"'
                                ', {}').format(option.impl_get_display_name(), err), err) from err
        except ValueError as err:
            raise ValueError(_('the option "{0}" is used in a calculation but is invalid ({1})').format(option.impl_get_display_name(), err)) from err
        except AttributeError as err:
            if isinstance(param, ParamDynOption) and param.optional:
                # cannot acces, simulate a propertyerror
                raise PropertiesOptionError(param,
                                            ['configerror'],
                                            config_bag.context.get_settings(),
                                            )
            raise ConfigError(_(f'unable to get value for calculating "{option.impl_get_display_name()}", {err}')) from err
        if len(options_bag) > 1:
            parent_option_bag = options_bag[-2]
        else:
            parent_option_bag = None
        return parent_option_bag, options_bag[-1]

    if isinstance(param, ParamValue):
        return param.value

    if isinstance(param, ParamInformation):
        if isinstance(param, ParamSelfInformation):
            option_bag = OptionBag(option,
                                   index,
                                   config_bag,
                                   )
        elif param.option:
            option_bag = OptionBag(param.option,
                                   None,
                                   config_bag,
                                   )
        else:
            option_bag = None
        try:
            return config_bag.context.get_values().get_information(option_bag,
                                                                   param.information_name,
                                                                   param.default_value,
                                                                   )
        except ValueError as err:
            raise ConfigError(_('option "{}" cannot be calculated: {}').format(option.impl_get_display_name(),
                                                                               str(err),
                                                                               ))

    if isinstance(param, ParamIndex):
        return index

    if isinstance(param, ParamSuffix):
        if not option.issubdyn():
            raise ConfigError(_('option "{}" is not in a dynoptiondescription').format(option.impl_get_display_name()))
        return option.impl_getsuffix()

    if isinstance(param, ParamSelfOption):
        value = calc_self(param,
                          option,
                          index,
                          orig_value,
                          config_bag,
                          )
        if callback.__name__ not in FUNCTION_WAITING_FOR_DICT:
            return value
        return {'name': option.impl_get_display_name(),
                'value': value,
                }

    if isinstance(param, ParamOption):
        callbk_option = param.option
        callbk_options = None
        if callbk_option.issubdyn():
            found = False
            if isinstance(param, ParamDynOption):
                od_path = param.dynoptiondescription.impl_getpath()
                if "." in od_path:
                    rootpath = od_path.rsplit('.', 1)[0] + '.'
                else:
                    rootpath = ''
                full_path = rootpath + param.subpath
                root_option_bag = OptionBag(config_bag.context.get_description(),
                                            None,
                                            config_bag,
                                            )
                try:
                    soptions_bag = config_bag.context.get_sub_option_bag(root_option_bag,
                                                                         full_path,
                                                                         #FIXME index?
                                                                         index=None,
                                                                         validate_properties=True,
                                                                         properties=None,
                                                                         )
                except AttributeError as err:
                    raise ConfigError(_(f'option "{option.impl_get_display_name()}" is not in a dynoptiondescription: {err}'))
                callbk_option = soptions_bag[-1].option
                found = True
            elif option.impl_is_sub_dyn_optiondescription():
                if option.getsubdyn() == callbk_option.getsubdyn():
                    root_path = option.impl_getpath().rsplit('.', 1)[0]
                    len_path = root_path.count('.')
                    full_path = root_path + '.' + callbk_option.impl_getpath().split('.', len_path + 1)[-1]
                    root_option_bag = OptionBag(config_bag.context.get_description(),
                                                None,
                                                config_bag,
                                                )
                    try:
                        soptions_bag = config_bag.context.get_sub_option_bag(root_option_bag,
                                                                             full_path,
                                                                             #FIXME index?
                                                                             index=None,
                                                                             validate_properties=True,
                                                                             properties=None,
                                                                             )
                    except AttributeError as err:
                        raise ConfigError(_(f'option "{option.impl_get_display_name()}" is not in a dynoptiondescription: {err}'))
                    callbk_option = soptions_bag[-1].option
                    found = True
            elif option.impl_is_dynsymlinkoption():
                rootpath = option.rootpath
                call_path = callbk_option.impl_getpath()
                if option.opt.issubdyn() and callbk_option.getsubdyn() == option.getsubdyn() or \
                        not option.opt.issubdyn() and callbk_option.getsubdyn() == option.opt:
                    # in same dynoption
                    suffix = option.impl_getsuffix()
                    subdyn = callbk_option.getsubdyn()
                    root_path, sub_path = subdyn.split_path(subdyn,
                                                            option,
                                                            )
                    if root_path:
                        parent_path = root_path + subdyn.impl_getname(suffix) + sub_path
                    else:
                        parent_path = subdyn.impl_getname(suffix) + sub_path
                    callbk_option = callbk_option.to_dynoption(parent_path,
                                                               suffix,
                                                               subdyn,
                                                               )
                    found = True
            if not found:
                callbk_options = []
                for doption_bag in callbk_option.getsubdyn().get_sub_children(callbk_option,
                                                                              config_bag,
                                                                              index=None,
                                                                              ):
                    callbk_options.append(doption_bag.option)
        if callbk_options is None:
            callbk_options = [callbk_option]
            values = None
        else:
            values = []
        for callbk_option in callbk_options:
            if index is not None and callbk_option.impl_get_leadership() and \
                    callbk_option.impl_get_leadership().in_same_leadership(option):
                if not callbk_option.impl_is_follower():
                    # leader
                    index_ = None
                    with_index = True
                else:
                    # follower
                    index_ = index
                    with_index = False
            else:
                index_ = None
                with_index = False
            if callbk_option.impl_getpath() == 'od.dodval1.st.boolean':
                raise Exception('pfff')
            value = get_value(config_bag,
                              callbk_option,
                              param,
                              index_,
                              False,
                              )
            if with_index:
                value = value[index]
            if values is not None:
                values.append(value)
        if values is not None:
            value = values
        if callback.__name__ not in FUNCTION_WAITING_FOR_DICT:
            return value
        return {'name': callbk_option.impl_get_display_name(),
                'value': value}


def carry_out_calculation(option,
                          callback: Callable,
                          callback_params: Optional[Params],
                          index: Optional[int],
                          config_bag: Optional[ConfigBag],
                          orig_value=undefined,
                          allow_value_error: bool=False,
                          force_value_warning: bool=False,
                          for_settings: bool=False,
                          ):
    """a function that carries out a calculation for an option's value

    :param option: the option
    :param callback: the name of the callback function
    :param callback_params: the callback's parameters
                            (only keyword parameters are allowed)
    :param index: if an option is multi, only calculates the nth value
    :param allow_value_error: to know if carry_out_calculation can return ValueError or ValueWarning (for example if it's a validation)
    :param force_value_warning: transform valueError to ValueWarning object

    The callback_params is a dict. Key is used to build args (if key is '')
    and kwargs (otherwise). Values are tuple of:
    - values
    - tuple with option and boolean's force_permissive (True when don't raise
    if PropertiesOptionError)
    Values could have multiple values only when key is ''."""
    if not option.impl_is_optiondescription() and option.impl_is_follower() and index is None:
        raise Exception('follower must have index in carry_out_calculation!')
    def fake_items(iterator):
        return ((None, i) for i in iterator)
    args = []
    kwargs = {}
    if callback_params:
        for key, param in chain(fake_items(callback_params.args), callback_params.kwargs.items()):
            try:
                value = manager_callback(callback,
                                         param,
                                         option,
                                         index,
                                         orig_value,
                                         config_bag,
                                         for_settings,
                                         )
                if key is None:
                    args.append(value)
                else:
                    kwargs[key] = value
            except PropertiesOptionError as err:
                if param.raisepropertyerror:
                    raise err
                if callback.__name__ in FUNCTION_WAITING_FOR_DICT:
                    if key is None:
                        args.append({'propertyerror': str(err), 'name': option.impl_get_display_name()})
                    else:
                        kwargs[key] = {'propertyerror': str(err), 'name': option.impl_get_display_name()}
    ret = calculate(option,
                    callback,
                    allow_value_error,
                    force_value_warning,
                    args,
                    kwargs,
                    )
    if isinstance(ret, list) and not option.impl_is_dynoptiondescription() and \
            option.impl_is_follower() and not option.impl_is_submulti():
        if args or kwargs:
            raise LeadershipError(_('the "{}" function with positional arguments "{}" '
                                    'and keyword arguments "{}" must not return '
                                    'a list ("{}") for the follower option "{}"'
                                    '').format(callback.__name__,
                                               args,
                                               kwargs,
                                               ret,
                                               option.impl_get_display_name()))
        else:
            raise LeadershipError(_('the "{}" function must not return a list ("{}") '
                                    'for the follower option "{}"'
                                    '').format(callback.__name__,
                                               ret,
                                               option.impl_get_display_name()))
    return ret


def calculate(option,
              callback: Callable,
              allow_value_error: bool,
              force_value_warning: bool,
              args,
              kwargs,
              ):
    """wrapper that launches the 'callback'

    :param callback: callback function
    :param args: in the callback's arity, the unnamed parameters
    :param kwargs: in the callback's arity, the named parameters

    """
    try:
        return callback(*args, **kwargs)
    except (ValueError, ValueWarning) as err:
        if allow_value_error:
            if force_value_warning:
                raise ValueWarning(str(err))
            raise err
        error = err
    except Exception as err:
        error = err
    if args or kwargs:
        msg = _('unexpected error "{0}" in function "{1}" with arguments "{3}" and "{4}" '
                'for option "{2}"').format(str(error),
                                           callback.__name__,
                                           option.impl_get_display_name(),
                                           args,
                                           kwargs)
    else:
        msg = _('unexpected error "{0}" in function "{1}" for option "{2}"'
                '').format(str(error),
                           callback.__name__,
                           option.impl_get_display_name())
    raise ConfigError(msg) from error



# -*- coding: UTF-8 -*-
# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough gus of pypy: pypy: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
"internationalisation utilities"
from .log import log

from gettext import translation, NullTranslations
from platform import system
from pkg_resources import resource_filename
from os import environ


DEFAULT = 'en'


def get_translation() -> str:
    """Sets the user locale as langage
    The default is set to english
    """
    # Application name (without .i18n)
    app_name = __name__[:-5]
    translations_path = resource_filename(app_name, 'locale')

    if 'TIRAMISU_LOCALE' in environ:  # pragma: no cover
        user_locale = environ['TIRAMISU_LOCALE']
    else:
        if 'Windows' in system():  # pragma: no cover
            import ctypes
            from locale import windows_locale
            default_locale = windows_locale[ctypes.windll.kernel32.GetUserDefaultUILanguage()]
        else:
            from locale import getlocale
            default_locale = getlocale()
        if default_locale and isinstance(default_locale, tuple):
            if default_locale[0] is not None:
                user_locale = default_locale[0][:2]
            else:
                user_locale = DEFAULT
        elif default_locale:  # pragma: no cover
            user_locale = default_locale[:2]
        else:  # pragma: no cover
            user_locale = DEFAULT
    try:
        trans = translation(domain=app_name,
                            localedir=translations_path,
                            languages=[user_locale],
                            )
#                            codeset='UTF-8')
    except FileNotFoundError:  # pragma: no cover
        log.debug('cannot found translation file for langage {} in localedir {}'.format(user_locale,
                                                                                        translations_path))
        trans = NullTranslations()
    return trans.gettext


_ = get_translation()



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""UsernameOption or GroupnameOption to check unix username/group value
"""
import re

from ..i18n import _
from .stroption import RegexpOption


class UsernameOption(RegexpOption):
    """UsernameOption to check unix username value
    """
    __slots__ = tuple()
    #regexp build with 'man 8 adduser' informations
    _regexp = re.compile(r"^[a-z_][a-z0-9_-]{0,30}[$a-z0-9_-]{0,1}$")
    _type = _('unix username')


class GroupnameOption(UsernameOption):
    """GroupnameOption to check unix group value
    """
    __slots__ = tuple()
    _type = _('unix groupname')



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
"user defined exceptions"
import weakref
from .i18n import _


def display_list(lst, separator='and', add_quote=False):
    if not lst:
        return '""'
    if separator == 'and':
        separator = _('and')
    elif separator == 'or':
        separator = _('or')
    if isinstance(lst, tuple) or isinstance(lst, frozenset):
        lst = list(lst)
    if len(lst) == 1:
        ret = lst[0]
        if not isinstance(ret, str):
            ret = str(ret)
        if add_quote and not ret.startswith('"'):
            ret = '"{}"'.format(ret)
        return ret
    lst_ = []
    for l in lst:
        if not isinstance(l, str):
            l = str(l)
        lst_.append(_(l))
    lst__ = []
    for l in lst_:
        if add_quote and not l.startswith('"'):
            l = '"{}"'.format(l)
        lst__.append(l)
    lst__.sort()
    last = lst__[-1]
    return ', '.join(lst__[:-1]) + _(' {} ').format(separator) + '{}'.format(last)


# Exceptions for an Option
class PropertiesOptionError(AttributeError):
    "attempt to access to an option with a property that is not allowed"
    def __init__(self,
                 option_bag,
                 proptype,
                 settings,
                 opt_type=None,
                 name=None,
                 orig_opt=None,
                 help_properties=None):
        if opt_type:
            self._opt_type = opt_type
            self._name = name
            self._orig_opt = orig_opt
        else:
            if option_bag.option.impl_is_optiondescription():
                self._opt_type = 'optiondescription'
            else:
                self._opt_type = 'option'
            self._name = option_bag.option.impl_get_display_name()
            self._orig_opt = None
        self._option_bag = option_bag
        self.proptype = proptype
        self.help_properties = help_properties
        self._settings = settings
        self.msg = None
        super().__init__(None)

    def set_orig_opt(self, opt):
        self._orig_opt = opt

    def __str__(self):
        # this part is a bit slow, so only execute when display
        if self.msg is not None:
            return self.msg
        if self._settings is None:
            return 'error'
        if self.help_properties:
            properties = list(self.help_properties)
        else:
            properties = list(self.proptype)
        only_one = len(properties) == 1
        properties_msg = display_list(properties, add_quote=True)
        if only_one:
            prop_msg = _('property')
        else:
            prop_msg = _('properties')
        if properties == ['frozen']:
            if self._orig_opt:
                msg = 'cannot modify the {0} "{1}" because "{2}" has {3} {4}'
            else:
                msg = 'cannot modify the {0} "{1}" because has {2} {3}'
        else:
            if self._orig_opt:
                msg = 'cannot access to {0} "{1}" because "{2}" has {3} {4}'
            else:
                msg = 'cannot access to {0} "{1}" because has {2} {3}'
        if self._orig_opt:
            self.msg = _(msg).format(self._opt_type,
                                     self._orig_opt.impl_get_display_name(),
                                     self._name,
                                     prop_msg,
                                     properties_msg)
        else:
            self.msg = _(msg).format(self._opt_type,
                                     self._name,
                                     prop_msg,
                                     properties_msg)
        del self._opt_type, self._name
        del self._settings, self._orig_opt
        return self.msg


#____________________________________________________________
# Exceptions for a Config
class ConfigError(Exception):
    """attempt to change an option's owner without a value
    or in case of `_descr` is None
    or if a calculation cannot be carried out"""
    def __init__(self,
                 exp,
                 ori_err=None,
                 ):
        super().__init__(exp)
        self.ori_err = ori_err


class ConflictError(Exception):
    "duplicate options are present in a single config"
    pass


#____________________________________________________________
# miscellaneous exceptions
class LeadershipError(Exception):
    "problem with a leadership's value length"
    pass


class ConstError(TypeError):
    "no uniq value in _NameSpace"
    pass


class _CommonError:
    def __init__(self,
                 val,
                 display_type,
                 opt,
                 err_msg,
                 index):
        self.val = val
        self.display_type = display_type
        self.opt = weakref.ref(opt)
        self.name = opt.impl_get_display_name()
        self.err_msg = err_msg
        self.index = index
        super().__init__(self.err_msg)

    def __str__(self):
        try:
            msg = self.prefix
        except AttributeError:
            self.prefix = self.tmpl.format(self.val,
                                           self.display_type,
                                           self.name)
            msg = self.prefix
        if self.err_msg:
            if msg:
                msg += ', {}'.format(self.err_msg)
            else:
                msg = self.err_msg
        if not msg:
            msg = _('invalid value')
        return msg


class ValueWarning(_CommonError, UserWarning):
    tmpl = _('attention, "{0}" could be an invalid {1} for "{2}"')

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            self.msg = args[0]
        else:
            super().__init__(*args, **kwargs)
            self.msg = None

    def __str__(self):
        if self.msg is None:
            return super().__str__()
        return self.msg


class ValueOptionError(_CommonError, ValueError):
    tmpl = _('"{0}" is an invalid {1} for "{2}"')


class ValueErrorWarning(ValueWarning):
    tmpl = _('"{0}" is an invalid {1} for "{2}"')



# -*- coding: utf-8 -*-
# Copyright (C) 2014-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""base option
"""
from typing import FrozenSet, Set, Any, List
import weakref
from itertools import chain


from ..i18n import _
from ..setting import undefined
from ..autolib import Calculation, ParamOption

STATIC_TUPLE = frozenset()


submulti = 2


def valid_name(name):
    """valid option name
    """
    if not isinstance(name, str):
        return False
#    if '.' in name:
#        return False
    return True


#____________________________________________________________
#
class Base:
    """Base use by all *Option* classes (Option, OptionDescription, SymLinkOption, ...)
    """
    __slots__ = ('_name',
                 '_path',
                 '_informations',
                 '_subdyns',
                 '_properties',
                 '_has_dependency',
                 '_dependencies',
                 '_suffixes_dependencies',
                 '__weakref__'
                )

    def __init__(self,
                 name: str,
                 doc: str,
                 properties=None,
                 is_multi: bool=False) -> None:
        if not valid_name(name):
            raise ValueError(_('"{0}" is an invalid name for an option').format(name))
        if properties is None:
            properties = frozenset()
        elif isinstance(properties, tuple):
            properties = frozenset(properties)
        if is_multi:
            # if option is a multi, it cannot be 'empty' (None not allowed in the list)
            # and cannot have multiple time the same value
            # 'empty' and 'unique' are removed for follower's option
            if 'notunique' not in properties:
                properties = properties | {'unique'}
            if 'notempty' not in properties:
                properties = properties | {'empty'}
        assert isinstance(properties, frozenset), _('invalid properties type {0} for {1},'
                                                    ' must be a frozenset').format(type(properties),
                                                                                   name)
        for prop in properties:
            if not isinstance(prop, str):
                if not isinstance(prop, Calculation):
                    raise ValueError(_('invalid property type {0} for {1}, must be a string or a '
                                       'Calculation').format(type(prop), name))
                for param in chain(prop.params.args, prop.params.kwargs.values()):
                    if isinstance(param, ParamOption):
                        param.option._add_dependency(self)
        _setattr = object.__setattr__
        _setattr(self, '_name', name)
        _setattr(self, '_informations', {'doc': doc})
        if properties:
            _setattr(self, '_properties', properties)

    def impl_has_dependency(self,
                            self_is_dep: bool=True,
                            ) -> bool:
        """this has dependency
        """
        if self_is_dep is True:
            return getattr(self, '_has_dependency', False)
        return hasattr(self, '_dependencies')

    def get_dependencies(self,
                         context_od,
                         ) -> Set[str]:
        ret = set(getattr(self, '_dependencies', STATIC_TUPLE))
        if context_od and hasattr(context_od, '_dependencies'):
            # add options that have context is set in calculation
            return set(context_od._dependencies) | ret  # pylint: disable=protected-access
        return ret

    def _get_suffixes_dependencies(self) -> Set[str]:
        return getattr(self, '_suffixes_dependencies', STATIC_TUPLE)

    def _add_dependency(self,
                        option,
                        is_suffix: bool=False,
                        ) -> None:
        woption = weakref.ref(option)
        options = self.get_dependencies(None)
        options.add(woption)
        self._dependencies = tuple(options)  # pylint: disable=attribute-defined-outside-init
        if is_suffix:
            options = list(self._get_suffixes_dependencies())
            options.append(woption)
            self._suffixes_dependencies = tuple(options)  # pylint: disable=attribute-defined-outside-init

    def impl_is_optiondescription(self) -> bool:
        """option is an option description
        """
        return False

    def impl_is_dynoptiondescription(self) -> bool:
        """option is not a dyn option description
        """
        return False

    def impl_is_sub_dyn_optiondescription(self):
        return False

    def impl_getname(self) -> str:
        """get name
        """
        return self._name  # pylint: disable=no-member

    def _set_readonly(self) -> None:
        if isinstance(self._informations, dict):  # pylint: disable=no-member
            _setattr = object.__setattr__
            dico = self._informations  # pylint: disable=no-member
            keys = tuple(dico.keys())
            if len(keys) == 1:
                dico = dico['doc']
            else:
                dico = tuple([keys, tuple(dico.values())])
            _setattr(self, '_informations', dico)
            extra = getattr(self, '_extra', None)
            if extra is not None:
                _setattr(self, '_extra', tuple([tuple(extra.keys()), tuple(extra.values())]))

    def impl_is_readonly(self) -> str:
        """the option is readonly
        """
        # _path is None when initialise SymLinkOption
        return hasattr(self, '_path') and self._path is not None  # pylint: disable=no-member

    def impl_getproperties(self) -> FrozenSet[str]:
        """get properties
        """
        return getattr(self, '_properties', frozenset())

    def _setsubdyn(self,
                   subdyn,
                   ) -> None:
        # pylint: disable=attribute-defined-outside-init
        if getattr(self, '_subdyns', None) is None:
            self._subdyns = []
        self._subdyns.append(subdyn)

    def issubdyn(self) -> bool:
        """is sub dynoption
        """
        return getattr(self, '_subdyns', None) is not None

    def getsubdyn(self):
        """get sub dynoption
        """
        return self._subdyns[0]()

    def get_sub_dyns(self):
        return self._subdyns

    # ____________________________________________________________
    # information
    def impl_get_information(self,
                             key: str,
                             default: Any=undefined,
                             ) -> Any:
        """retrieves one information's item

        :param key: the item string (ex: "help")
        """
        dico = self._informations  # pylint: disable=no-member
        if isinstance(dico, tuple):
            if key in dico[0]:
                return dico[1][dico[0].index(key)]
        elif isinstance(dico, str):
            if key == 'doc':
                return dico
        elif isinstance(dico, dict):
            if key in dico:
                return dico[key]
        if default is not undefined:
            return default
        # pylint: disable=no-member
        raise ValueError(_(f'information\'s item for "{self.impl_get_display_name()}" '
                           f'not found: "{key}"'))

    def impl_set_information(self,
                             key: str,
                             value: Any,
                             ) -> None:
        """updates the information's attribute
        (which is a dictionary)

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        if self.impl_is_readonly():
            raise AttributeError(_("'{0}' ({1}) object attribute '{2}' is"
                                   " read-only").format(self.__class__.__name__,
                                                        self,
                                                        key))
        self._informations[key] = value  # pylint: disable=no-member

    def impl_list_information(self) -> Any:
        """get the list of information keys
        """
        dico = self._informations  # pylint: disable=no-member
        if isinstance(dico, tuple):
            return list(dico[0])
        if isinstance(dico, str):
            return ['doc']
        # it's a dict
        return list(dico.keys())


class BaseOption(Base):
    """This abstract base class stands for attribute access
    in options that have to be set only once, it is of course done in the
    __setattr__ method
    """
    __slots__ = ('_display_name_function',)

    def __setattr__(self,
                    name: str,
                    value: Any,
                    ) -> Any:
        """set once and only once some attributes in the option,
        like `_name`. `_name` cannot be changed once the option is
        pushed in the :class:`tiramisu.option.OptionDescription`.

        if the attribute `_readonly` is set to `True`, the option is
        "frozen" (which has nothing to do with the high level "freeze"
        propertie or "read_only" property)
        """
        # never change _name in an option or attribute when object is readonly
        if self.impl_is_readonly():
            raise AttributeError(_('"{}" ({}) object attribute "{}" is'
                                   ' read-only').format(self.__class__.__name__,
                                                        self.impl_get_display_name(),
                                                        name))
        super().__setattr__(name, value)

    def impl_getpath(self) -> str:
        """get the path of the option
        """
        try:
            return self._path
        except AttributeError as err:
            raise AttributeError(_(f'"{self.impl_get_display_name()}" not part of any Config')) \
                    from err

    def impl_get_display_name(self,
                              dynopt=None,
                              ) -> str:
        """get display name
        """
        if dynopt is None:
            dynopt = self
        if hasattr(self, '_display_name_function'):
            return self._display_name_function(dynopt)
        name = self.impl_get_information('doc', None)
        if name is None or name == '':
            name = dynopt.impl_getname()
        return name

    def reset_cache(self,
                    path: str,
                    config_bag: 'OptionBag',
                    resetted_opts: List[Base],  # pylint: disable=unused-argument
                    ) -> None:
        """reset cache
        """
        context = config_bag.context
        context.properties_cache.delcache(path)
        context._impl_permissives_cache.delcache(path)  # pylint: disable=protected-access
        if not self.impl_is_optiondescription():
            context.get_values_cache().delcache(path)  # pylint: disable=protected-access

    def impl_is_symlinkoption(self) -> bool:
        """the option is not a symlinkoption
        """
        return False

    def get_dependencies_information(self) -> List[str]:
        """get dependencies information
        """
        return getattr(self, '_dependencies_information', {})



# -*- coding: utf-8 -*-

import warnings
import sys
from copy import copy
from itertools import chain
from .error import ValueWarning, ValueErrorWarning, PropertiesOptionError, ConfigError
from .setting import undefined
from . import SynDynOption, RegexpOption, ChoiceOption, ParamOption
from .i18n import _


CONVERT_WEB_TYPE = {'bool': 'boolean',
                    'str': 'string',
                    'int': 'integer',
                    }
INPUTS = ['string',
          'integer',
          'filename',
          'password',
          'email',
          'username',
          'ip',
          'domainname']


# return always warning (even if same warning is already returned)
warnings.simplefilter("always", ValueWarning)
warnings.simplefilter("always", ValueErrorWarning)


class Callbacks(object):
    def __init__(self, tiramisu_web):
        self.tiramisu_web = tiramisu_web
        self.clearable = tiramisu_web.clearable
        self.remotable = tiramisu_web.remotable
        self.callbacks = []

    def add(self,
            path,
            childapi,
            schema,
            force_store_value):
        if self.remotable == 'all' or childapi.isoptiondescription():
            return
#        callback, callback_params = childapi.callbacks()
#        if callback is None:  # FIXME ? and force_store_value and self.clearable != 'all':
#            return
#        self.callbacks.append((callback, callback_params, path, childapi, schema, force_store_value))

    def process_properties(self, form):
        for callback, callback_params, path, childapi, schema, force_store_value in self.callbacks:
            if childapi.isfollower():
                self.tiramisu_web.set_remotable(path, form, childapi)
                continue
            has_option = False
            if callback_params is not None:
                for callback_param in chain(callback_params.args, callback_params.kwargs.values()):
                    if isinstance(callback_param, ParamOption):
                        has_option = True
                        if 'expire' in childapi.properties():
                            self.tiramisu_web.set_remotable(callback_param.impl_getpath(), form)
            if not has_option and form.get(path, {}).get('remote', False) == False:
                if 'expire' in childapi.properties():
                    self.tiramisu_web.set_remotable(path, form, childapi)
                elif childapi.owner.isdefault():
                    # get calculated value and set clearable
                    schema[path]['value'] = childapi.value.get()
                    if self.clearable == 'minimum':
                        form.setdefault(path, {})['clearable'] = True

    def manage_callbacks(self, form):
        pass
        #for callback, callback_params, path, childapi, schema, force_store_value in self.callbacks:
        #    if callback_params is not None:
        #        for callback_param in chain(callback_params.args, callback_params.kwargs.values()):
        #            if isinstance(callback_param, ParamOption) and callback.__name__ == 'tiramisu_copy':
        #                opt_path = callback_param.impl_getpath()
        #                if form.get(opt_path, {}).get('remote') is not True:
        #                    form.setdefault(opt_path, {})
        #                    form[opt_path].setdefault('copy', []).append(path)

    def process(self,
                      form):
        self.process_properties(form)
        self.manage_callbacks(form)


class Consistencies(object):
    def __init__(self, tiramisu_web):
        self.not_equal = {}
        self.tiramisu_web = tiramisu_web

    def add(self, path, childapi, form):
        return
        if not childapi.isoptiondescription():
            for consistency in childapi.consistencies():
                cons_id, func, all_cons_opts, params = consistency
                if func == '_cons_not_equal' and params.get('transitive', True) is True:
                    options_path = []
                    for option in all_cons_opts:
                        options_path.append(option()._path)
                    for idx, option in enumerate(all_cons_opts):
                        option = option()
                        paths = options_path.copy()
                        paths.pop(idx)
                        warnings_only = params.get('warnings_only') or getattr(option, '_warnings_only', False)
                        self.not_equal.setdefault(option._path, {}).setdefault(warnings_only, []).extend(paths)
                else:
                    for option in all_cons_opts:
                        self.tiramisu_web.set_remotable(option()._path, form)

    def process(self, form):
        for path in self.not_equal:
            if self.tiramisu_web.is_remote(path, form):
                continue
            if path not in form:
                form[path] = {}
            for warnings_only in self.not_equal[path]:
                options = self.not_equal[path][warnings_only]
                if 'not_equal' not in form[path]:
                    form[path]['not_equal'] = []
                obj = {'options': options}
                if warnings_only:
                    obj['warnings'] = True
                form[path]['not_equal'].append(obj)


class Requires(object):
    def __init__(self, tiramisu_web):
        self.requires = {}
        self.options = {}
        self.tiramisu_web = tiramisu_web
        self.action_hide = self.tiramisu_web.config._config_bag.properties

    def set_master_remote(self, childapi, path, form):
        if childapi.isoptiondescription():
            isfollower = False
        else:
            isfollower = childapi.isfollower()
        if isfollower:
            parent_path = path.rsplit('.', 1)[0]
            parent = self.tiramisu_web.config.unrestraint.option(parent_path)
            leader = parent.list()[0]
            self.tiramisu_web.set_remotable(leader.path(), form, leader)

    def manage_requires(self,
                        childapi,
                        path,
                        form,
                        current_action,
                        ):
        for requires in childapi.properties(uncalculated=True):
            if not isinstance(requires, str):
                option = requires.params.kwargs['condition'].option
                expected = [requires.params.kwargs['expected'].value]
                action = requires.params.args[0].value
                if 'reverse_condition' in requires.params.kwargs:
                    inverse = requires.params.kwargs['reverse_condition'].value
                else:
                    inverse = False
                transitive = True
                same_action = True
                operator = 'or'
                if 1 == 1:
#            len_to_long = len(requires) > 1
#            for require in requires:
#                options, action, inverse, transitive, same_action, operator = require
#                if not len_to_long:
#                    len_to_long = len(options) > 1
#                for option, expected in options:
                    if isinstance(option, tuple):
                        for option_param in chain(option[1].args, option[1].kwargs.values()):
                            if isinstance(option_param, ParamOption):
                                self.tiramisu_web.set_remotable(option_param.option.impl_getpath(), form)
                        self.set_master_remote(childapi, path, form)
#                    elif len_to_long:
#                        self.tiramisu_web.set_remotable(option.impl_getpath(), form)
#                        self.set_master_remote(childapi, path, form)
                    else:
                        option_path = option.impl_getpath()
                        if action in self.action_hide:
                            require_option = self.tiramisu_web.config.unrestraint.option(option_path)
                            if transitive is False or same_action is False or operator == 'and':
                                # transitive to "False" not supported yet for a requirement
                                # same_action to "False" not supported yet for a requirement
                                # operator "and" not supported yet for a requirement
                                self.tiramisu_web.set_remotable(option_path, form, require_option)
                                self.set_master_remote(childapi, path, form)
#                            if require_option.option.requires():
#                                for reqs in require_option.option.requires():
#                                    for req in reqs:
#                                        for subopt, subexp in req[0]:
#                                            if not isinstance(subopt, tuple):
#                                                self.tiramisu_web.set_remotable(subopt.impl_getpath(), form)
#                                                self.set_master_remote(childapi, path, form)
                            if inverse:
                                act = 'show'
                                inv_act = 'hide'
                            else:
                                act = 'hide'
                                inv_act = 'show'
                            if isinstance(option, ChoiceOption):
                                require_option = self.tiramisu_web.config.unrestraint.option(option_path)
                                values = self.tiramisu_web.get_enum(require_option,
                                                                          require_option.ismulti(),
                                                                          option_path,
                                                                          require_option.properties())
                                for value in values:
                                    if value not in expected:
                                        self.requires.setdefault(path,
                                                                  {'expected': {}}
                                                                 )['expected'].setdefault(value,
                                                                                          {}).setdefault(inv_act,
                                                                                                         []).append(option_path)
                            if current_action is None:
                                current_action = action
                            elif current_action != action:
                                self.tiramisu_web.set_remotable(option_path, form)
                                self.set_master_remote(childapi, path, form)
                            for exp in expected:
                                self.requires.setdefault(path,
                                                         {'expected': {}}
                                                         )['expected'].setdefault(exp,
                                                                                  {}).setdefault(act,
                                                                                                 []).append(option_path)
                            self.requires[path].setdefault('default', {}).setdefault(inv_act, []).append(option_path)
                        else:
                            self.tiramisu_web.set_remotable(option_path, form)
                            self.set_master_remote(childapi, path, form)

    def add(self, path, childapi, form):
        #collect id of all options
        child = childapi.get()
        if isinstance(child, SynDynOption):
            child = child.opt
        self.options[child] = path
        current_action = None

        self.manage_requires(childapi,
                             path,
                             form,
                             current_action,
                             )

    def process(self, form):
        dependencies = {}
        for path, values in self.requires.items():
            if 'default' in values:
                for option in values['default'].get('show', []):
                    if path == option:
                        self.tiramisu_web.set_remotable(path, form)
                    if not self.tiramisu_web.is_remote(option, form):
                        dependencies.setdefault(option,
                                                {'default': {}, 'expected': {}}
                                                )['default'].setdefault('show', [])
                        if path not in dependencies[option]['default']['show']:
                            dependencies[option]['default']['show'].append(path)
                for option in values['default'].get('hide', []):
                    if path == option:
                        self.tiramisu_web.set_remotable(path, form)
                    if not self.tiramisu_web.is_remote(option, form):
                        dependencies.setdefault(option,
                                                {'default': {}, 'expected': {}}
                                                )['default'].setdefault('hide', [])
                        if path not in dependencies[option]['default']['hide']:
                            dependencies[option]['default']['hide'].append(path)
            for expected, actions in values['expected'].items():
                if expected is None:
                    expected = ''
                for option in actions.get('show', []):
                    if path == option:
                        self.tiramisu_web.set_remotable(path, form)
                    if not self.tiramisu_web.is_remote(option, form):
                        dependencies.setdefault(option,
                                                {'expected': {}}
                                                )['expected'].setdefault(expected,
                                                                         {}).setdefault('show', [])
                        if path not in dependencies[option]['expected'][expected]['show']:
                            dependencies[option]['expected'][expected]['show'].append(path)
                for option in actions.get('hide', []):
                    if path == option:
                        self.tiramisu_web.set_remotable(path, form)
                    if not self.tiramisu_web.is_remote(option, form):
                        dependencies.setdefault(option,
                                                {'expected': {}}
                                                )['expected'].setdefault(expected,
                                                                         {}).setdefault('hide', [])
                        if path not in dependencies[option]['expected'][expected]['hide']:
                            dependencies[option]['expected'][expected]['hide'].append(path)
        for path, dependency in dependencies.items():
            form.setdefault(path, {})['dependencies'] = dependency


class TiramisuDict:

    # propriete:
    #   hidden
    #   mandatory
    #   editable

    # FIXME model:
    # #optionnel mais qui bouge
    # choices/suggests
    # warning
    #
    # #bouge
    # owner
    # properties

    def __init__(self,
                 config,
                 root=None,
                 clearable="all",
                 remotable="minimum"):
        self.config = config
        self.root = root
        self.requires = None
        self.callbacks = None
        self.consistencies = None
        #all, minimum, none
        self.clearable = clearable
        #all, minimum, none
        self.remotable = remotable

    def add_help(self,
                       obj,
                       childapi):
        hlp = childapi.information.get('help', None)
        if hlp is not None:
            obj['help'] = hlp

    def get_list(self, root, subchildapi):
        #ret = []
        for childapi in subchildapi.list('all'):
            childname = childapi.name()
            if root is None:
                path = childname
            else:
                path = root + '.' + childname
            yield (path, childapi)
        #return ret

    def is_remote(self, path, form):
        if self.remotable == 'all':
            return True
        else:
            return path in form and form[path].get('remote', False) == True

    def set_remotable(self, path, form, childapi=None):
        if self.remotable == 'none':
            raise ValueError(_('option {} only works when remotable is not "none"').format(path))
        form.setdefault(path, {})['remote'] = True
        if childapi is None:
            childapi = self.config.unrestraint.option(path)
        if childapi.isfollower():
            parent_path = path.rsplit('.', 1)[0]
            parent = self.config.unrestraint.option(parent_path)
            leader = parent.list()[0]
            form.setdefault(leader.path(), {})['remote'] = True

    def walk(self,
             root,
             subchildapi,
             schema,
             model,
             form,
             order,
             updates_status,
             init=False,
             ):
        error = None
        if init:
            if form is not None:
                self.requires = Requires(self)
                self.consistencies = Consistencies(self)
                self.callbacks = Callbacks(self)
        else:
            init = False
        try:
            if subchildapi is None:
                if root is None:
                    subchildapi = self.config.unrestraint.option
                else:
                    subchildapi = self.config.unrestraint.option(root)
                isleadership = False
            else:
                isleadership = subchildapi.isleadership()
            leader_len = None
            for path, childapi in self.get_list(root, subchildapi):
                if isleadership and leader_len is None:
                    leader_len = childapi.value.len()
                    one_is_remote = False
                if not childapi.isoptiondescription() and childapi.isfollower():
                    props_no_requires = set()
                else:
                    props_no_requires = set(childapi.properties())
                if form is not None:
                    self.requires.add(path,
                                      childapi,
                                      form,
                                      )
                    self.consistencies.add(path,
                                           childapi,
                                           form,
                                           )
                    self.callbacks.add(path,
                                       childapi,
                                       schema,
                                       'force_store_value' in props_no_requires,
                                       )
                if model is not None and childapi.isoptiondescription() or not childapi.issymlinkoption():
                    self.gen_model(model,
                                   childapi,
                                   path,
                                   leader_len,
                                   updates_status,
                                   )
                if order is not None:
                    order.append(path)
                if childapi.isoptiondescription():
                    web_type = 'optiondescription'
                    if childapi.isleadership():
                        type_ = 'array'
                    else:
                        type_ = 'object'
                    if schema is not None:
                        schema[path] = {'properties': {},
                                        'type': type_}
                        subschema = schema[path]['properties']
                    else:
                        subschema = schema
                    self.walk(path,
                              childapi,
                              subschema,
                              model,
                              form,
                              order,
                              updates_status,
                              )
                else:
                    child = childapi.get()
                    childtype = child.__class__.__name__
                    if childtype == 'SynDynOption':
                        childtype = child.opt.__class__.__name__
                    if childapi.issymlinkoption():
                        web_type = 'symlink'
                        value = None
                        defaultmulti = None
                        is_multi = False
                    else:
                        web_type = childapi.get().__class__.__name__.lower()[:-6]
                        web_type = CONVERT_WEB_TYPE.get(web_type, web_type)
                        value = childapi.default()
                        if value == []:
                            value = None

                        is_multi = childapi.ismulti()
                        if is_multi:
                            defaultmulti = childapi.defaultmulti()
                            if defaultmulti == []:
                                defaultmulti = None
                        else:
                            defaultmulti = None
                    if schema is not None:
                        self.gen_schema(schema,
                                        childapi,
                                        path,
                                        props_no_requires,
                                        value,
                                        defaultmulti,
                                        is_multi,
                                        web_type,
                                        form,
                                        )
                    if form is not None:
                        self.gen_form(form,
                                      web_type,
                                      path,
                                      child,
                                      childapi,
                                      childtype,
                                      )
                if schema is not None:
                    if web_type != 'symlink':
                        schema[path]['title'] = childapi.description()
                    self.add_help(schema[path],
                                        childapi)
        except Exception as err:
            import traceback
            traceback.print_exc()
            if not init:
                raise err
            error = err
        if init and form is not None:
            self.callbacks.process(form)
            self.requires.process(form)
            self.consistencies.process(form)
            del self.requires
            del self.consistencies
            del self.callbacks
        if error:
            msg = str(error)
            del error
            raise ConfigError(_('unable to transform tiramisu object to dict: {}').format(msg))


    def gen_schema(self,
                   schema,
                   childapi,
                   path,
                   props_no_requires,
                   value,
                   defaultmulti,
                   is_multi,
                   web_type,
                   form,
                   ):
        schema[path] = {'type': web_type}
        if childapi.issymlinkoption():
            sym_option = childapi.get()
            schema[path]['opt_path'] = sym_option.impl_getopt().impl_getpath()
        else:
            if defaultmulti is not None:
                schema[path]['defaultmulti'] = defaultmulti

            if is_multi:
                schema[path]['isMulti'] = is_multi

            if childapi.issubmulti():
                schema[path]['isSubMulti'] = True

            if 'auto_freeze' in props_no_requires:
                schema[path]['autoFreeze'] = True

            if web_type == 'choice':
                #values, values_params = childapi.value.callbacks()
                #if values_params:
                #    for values_param in chain(values_params.args, values_params.kwargs.values()):
                #        if isinstance(values_param, ParamOption):
                #            self.set_remotable(path, form, childapi)
                #            return
                schema[path]['enum'] = self.get_enum(childapi,
                                                           is_multi,
                                                           path,
                                                           props_no_requires)
            if value is not None and not self.is_remote(path, form):
                schema[path]['value'] = value


    def get_enum(self,
                 childapi,
                 is_multi,
                 path,
                 props_no_requires,
                 ):
        values = childapi.value.list()
        empty_is_required = not childapi.isfollower() and is_multi
        if '' not in values and ((empty_is_required and not 'empty' in props_no_requires) or \
                (not empty_is_required and not 'mandatory' in props_no_requires)):
            values = [''] + list(values)
        return values

    def gen_form(self,
                 form,
                 web_type,
                 path,
                 child,
                 childapi,
                 childtype,
                 ):
        obj_form = {}
        if path in form:
            obj_form.update(form[path])
        if not childapi.issymlinkoption():
            #if childapi_option.validator() != (None, None):
            #    obj_form['remote'] = True
            #    params = childapi_option.validator()[1]
            #    if params is not None:
            #        for param in chain(params.args, params.kwargs.values()):
            #            if isinstance(param, ParamOption):
            #                self.set_remotable(param.option.impl_getpath(), form)
            if self.clearable == 'all':
                obj_form['clearable'] = True
            if self.clearable != 'none':
                obj_form['clearable'] = True
            if self.remotable == 'all' or childapi.has_dependency():
                obj_form['remote'] = True
            if childtype == 'IPOption' and (child.impl_get_extra('_private_only') or not child.impl_get_extra('_allow_reserved') or child.impl_get_extra('_cidr')):
                obj_form['remote'] = True
            if childtype == 'DateOption':
                obj_form['remote'] = True
            if not obj_form.get('remote', False):
                pattern = childapi.pattern()
                if pattern is not None:
                    obj_form['pattern'] = pattern
                if childtype == 'PortOption':
                    obj_form['min'] = child.impl_get_extra('_min_value')
                    obj_form['max'] = child.impl_get_extra('_max_value')
            if childtype == 'FloatOption':
                obj_form['step'] = 'any'
            if web_type == 'choice':
                obj_form['type'] = 'choice'
            elif web_type in INPUTS:
                obj_form['type'] = 'input'
            if obj_form:
                form[path] = obj_form

    def calc_raises_properties(self,
                               obj,
                               childapi,
                               ):
        old_properties = childapi._config_bag.properties
        config = childapi._config_bag.context
        settings = config.get_settings()
        childapi._config_bag.properties = self.config.property.default('current')  # settings.get_context_properties(config._impl_properties_cache)
        childapi._config_bag.properties -= {'permissive'}
        properties = childapi.property.get(only_raises=True,
                                           uncalculated=True,
                                           )
        properties -= childapi.permissive.get()
        # 'hidden=True' means cannot access with or without permissive option
        # 'display=False' means cannot access only without permissive option
        if properties:
            obj['display'] = False
        properties -= self.config.permissive.get()
        if properties:
            obj['hidden'] = True
        childapi._config_bag.properties = old_properties

    def _gen_model_properties(self,
                              childapi,
                              path,
                              index,
                              ):
        isfollower = childapi.isfollower()
        props = set(childapi.property.get())
        obj = self.gen_properties(props,
                                  isfollower,
                                  childapi.ismulti(),
                                  index)
        self.calc_raises_properties(obj, childapi)
        return obj

    def gen_properties(self,
                       properties,
                       isfollower,
                       ismulti,
                       index):
        obj = {}
        if not isfollower and ismulti:
            if 'empty' in properties:
                if index is None:
                    obj['required'] = True
                properties.remove('empty')
            if 'mandatory' in properties:
                if index is None:
                    obj['needs_len'] = True
                properties.remove('mandatory')
        elif 'mandatory' in properties:
            if index is None:
                obj['required'] = True
            properties.remove('mandatory')
        if 'frozen' in properties:
            if index is None:
                obj['readOnly'] = True
            properties.remove('frozen')
        if 'hidden' in properties:
            properties.remove('hidden')
        if 'disabled' in properties:
            properties.remove('disabled')
        if properties:
            lprops = list(properties)
            lprops.sort()
            obj['properties'] = lprops
        return obj

    def gen_model(self,
                  model,
                  childapi,
                  path,
                  leader_len,
                  updates_status,
                  ):
        if childapi.isoptiondescription():
            props = set(childapi.property.get())
            obj = {}
            self.calc_raises_properties(obj, childapi)
            if props:
                lprops = list(props)
                lprops.sort()
                obj['properties'] = lprops
            try:
                self.config.option(path).get()
            except PropertiesOptionError:
                pass
        else:
            obj = self._gen_model_properties(childapi,
                                             path,
                                             None,
                                             )
            if childapi.isfollower():
                for index in range(leader_len):
                    follower_childapi = self.config.unrestraint.option(path, index)
                    sobj = self._gen_model_properties(follower_childapi,
                                                      path,
                                                      index,
                                                      )
                    self._get_model_value(follower_childapi,
                                          path,
                                          sobj,
                                          index,
                                          updates_status,
                                          )
                    if sobj:
                        model.setdefault(path, {})[str(index)] = sobj
            else:
                self._get_model_value(childapi,
                                      path,
                                      obj,
                                      None,
                                      updates_status,
                                      )
        if obj:
            if not childapi.isoptiondescription() and childapi.isfollower():
                model.setdefault(path, {})['null'] = obj
            else:
                model[path] = obj

    def _get_model_value(self,
                         childapi,
                         path,
                         obj,
                         index,
                         updates_status,
                         ):
        if path in updates_status and index in updates_status[path]:
            value = childapi.value.get()
            self._get_value_with_exception(obj,
                                           childapi,
                                           updates_status[path][index])
            del updates_status[path][index]
        else:
            try:
                with warnings.catch_warnings(record=True) as warns:
                    value = self.config.option(path, index=index).value.get()
                self._get_value_with_exception(obj,
                                               childapi,
                                               warns)
            except ValueError as err:
                self._get_value_with_exception(obj,
                                               childapi,
                                               [err])
                value = self.config.unrestraint.option(path, index=index).value.get()
            except PropertiesOptionError as err:
                config_bag = self.config._config_bag
                settings = config_bag.context.get_settings()
                if settings._calc_raises_properties(config_bag.properties,
                                                    config_bag.permissives,
                                                    set(err.proptype)):
                    obj['hidden'] = True
                obj['display'] = False
                value = childapi.value.get()
        if value is not None and value != []:
            obj['value'] = value
        if not childapi.owner.isdefault():
            obj['owner'] = childapi.owner.get()

    def _get_value_with_exception(self,
                                  obj,
                                  childapi,
                                  values):
        for value in values:
            if isinstance(value, ValueError):
                obj.setdefault('error', [])
                msg = str(value)
                if msg not in obj.get('error', []):
                    obj['error'].append(msg)
                    obj['invalid'] = True
            elif isinstance(value.message, ValueErrorWarning):
                value.message.prefix = ''
                obj.setdefault('error', [])
                msg = str(value.message)
                if msg not in obj.get('error', []):
                    obj['error'].append(msg)
                    obj['invalid'] = True
            else:
                value.message.prefix = ''
                obj.setdefault('warnings', [])
                msg = str(value.message)
                if msg not in obj.get('error', []):
                    obj['warnings'].append(msg)
                    obj['hasWarnings'] = True

    def gen_global(self):
        ret = {}
        ret['owner'] = self.config.owner.get()
        ret['properties'] = list(self.config.property.get())
        ret['properties'].sort()
        ret['permissives'] = list(self.config.permissive.get())
        ret['permissives'].sort()
        return ret

    def get_form(self, form):
        ret = []
        buttons = []
        dict_form = {}
        for form_ in form:
            if 'key' in form_:
                dict_form[form_['key']] = form_
            elif form_.get('type') == 'submit':
                if 'cmd' not in form_:
                    form_['cmd'] = 'submit'
                buttons.append(form_)
            else:
                raise ValueError(_('unknown form {}').format(form_))

        for key, form_ in self.form.items():
            form_['key'] = key
            if key in dict_form:
                form_.update(dict_form[key])
            ret.append(form_)
        ret.extend(buttons)
        return ret

    def del_value(self, childapi, path, index):
        if index is not None and childapi.isleader():
            childapi.value.pop(index)
        elif index is None or childapi.isfollower():
            childapi.value.reset()
        else:
            multi = childapi.value.get()
            multi.pop(index)
            childapi.value.set(multi)

    def add_value(self, childapi, path, value):
        multi = childapi.value.get()
        multi.append(value)
        childapi.value.set(multi)

    def mod_value(self, childapi, path, index, value):
        if index is None or childapi.isfollower():
            childapi.value.set(value)
        else:
            multi = childapi.value.get()
            if len(multi) < index + 1:
                multi.append(value)
            else:
                multi[index] = value
            childapi.value.set(multi)

    def apply_updates(self,
                      oripath,
                      updates,
                      model_ori,
                      ):
        updates_status = {}
        for update in updates:
            path = update['name']
            index = update.get('index')
            if oripath is not None and not path.startswith(oripath):
                raise ValueError(_('not in current area'))
            childapi = self.config.option(path)
            if childapi.isfollower():
                childapi = self.config.option(path, index)
            with warnings.catch_warnings(record=True) as warns:
                try:
                    if update['action'] == 'modify':
                        self.mod_value(childapi,
                                       path,
                                       index,
                                       update.get('value', undefined),
                                       )
                    elif update['action'] == 'delete':
                        self.del_value(childapi,
                                       path,
                                       index,
                                       )
                    elif update['action'] == 'add':
                        if childapi.ismulti():
                            self.add_value(childapi, path, update['value'])
                        else:
                            raise ValueError(_('only multi option can have action "add", but "{}" is not a multi').format(path))
                    else:
                        raise ValueError(_('unknown action {}').format(update['action']))
                except ValueError as err:
                    updates_status.setdefault(path, {})[index] = [err]
            if warns != []:
                updates_status.setdefault(path, {}).setdefault(index, []).extend(warns)
        return updates_status

    def set_updates(self,
                    body,
                    ):
        root_path = self.root
        updates = body.get('updates', [])
        updates_status = self.apply_updates(root_path,
                                            updates,
                                            body.get('model'),
                                            )
        if 'model' in body:
            order = []
            old_model = body['model']
            new_model = self.todict(order=order,
                                    build_schema=False,
                                    build_form=False,
                                    updates_status=updates_status,
                                    )
            values = {'updates': list_keys(old_model, new_model['model'], order, updates_status),
                      'model': new_model['model']}
        else:
            values = updates_status
        return values

    def todict(self,
               custom_form=[],
               build_schema=True,
               build_model=True,
               build_form=True,
               order=None,
               updates_status={},
               ):
        rootpath = self.root
        if build_schema:
            schema = {}
        else:
            schema = None
        if build_model:
            model = {}
        else:
            model = None
        if build_form:
            form = {}
            buttons = []
        else:
            form = None
        self.walk(rootpath,
                  None,
                  schema,
                  model,
                  form,
                  order,
                  updates_status,
                  init=True,
                  )
        if build_form:
            for form_ in custom_form:
                if 'key' in form_:
                    key = form_.pop('key')
                    form.setdefault(key, {}).update(form_)
                elif form_.get('type') == 'submit':
                    # FIXME if an Option has a key "null"?
                    form.setdefault(None, []).append(form_)
                else:
                    raise ValueError(_('unknown form {}').format(form_))
        ret = {}
        if build_schema:
            ret['schema'] = schema
        if build_model:
            ret['model'] = model
            ret['global'] = self.gen_global()
        if build_form:
            ret['form'] = form
        ret['version'] = '1.0'
        return ret


def list_keys(model_a, model_b, ordered_key, updates_status):
    model_a_dict = {}
    model_b_dict = {}

    keys_a = set(model_a.keys())
    keys_b = set(model_b.keys())

    keys = (keys_a ^ keys_b) | set(updates_status.keys())

    for key in keys_a & keys_b:
        keys_mod_a = set(model_a[key].keys())
        keys_mod_b = set(model_b[key].keys())
        if keys_mod_a != keys_mod_b:
            keys.add(key)
        else:
            for skey in keys_mod_a:
                if model_a[key][skey] != model_b[key][skey]:
                    keys.add(key)
                    break
    def sort_key(key):
        try:
            return ordered_key.index(key)
        except ValueError:
            return -1
    return sorted(list(keys), key=sort_key)



# -*- coding: utf-8 -*-
"takes care of the option's values and multi values"
# Copyright (C) 2013-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from typing import Union, Optional, List, Any
from .error import ConfigError
from .setting import owners, undefined, forbidden_owners, OptionBag
from .autolib import Calculation
from .i18n import _


class Values:
    """This class manage value (default value, stored value or calculated value
    It's also responsible of a caching utility.
    """
    # pylint: disable=too-many-public-methods
    __slots__ = ('_values',
                 '_informations',
                 '__weakref__',
                 )

    def __init__(self,
                 default_values: Union[None, dict]=None,
                 ) -> None:
        """
        Initializes the values's dict.

        :param default_values: values stored by default for this object

        """
        self._informations = {}
        # set default owner
        if not default_values:
            default_values = {None: {None: [None, owners.user]}}
        self._values = default_values

    #______________________________________________________________________
    # get value
    def get_cached_value(self,
                         option_bag: OptionBag,
                         ) -> Any:
        """get value directly in cache if set
        otherwise calculated value and set it in cache

        :returns: value
        """
        # try to retrive value in cache
        setting_properties = option_bag.config_bag.properties
        cache = option_bag.config_bag.context.get_values_cache()
        is_cached, value, validated = cache.getcache(option_bag,
                                                     'values',
                                                     )
        # no cached value so get value
        if not is_cached:
            value, has_calculation = self.get_value(option_bag)
        # validates and warns value
        if not validated:
            validate = option_bag.option.impl_validate(value,
                                                       option_bag,
                                                       check_error=True,
                                                       )
        if 'warnings' in setting_properties:
            option_bag.option.impl_validate(value,
                                            option_bag,
                                            check_error=False,
                                            )
        # set value to cache
        if not is_cached and not has_calculation:
            cache.setcache(option_bag,
                           value,
                           validated=validate,
                           )
        if isinstance(value, list):
            # return a copy, so value cannot be modified
            value = value.copy()
        # and return it
        return value

    def get_value(self,
                  option_bag: OptionBag,
                  ) -> Any:
        """actually retrieves the stored value or the default value (value modified by user)

        :returns: value
        """
        # get owner and value from store
        default_value = [undefined, owners.default]
        value, owner = self._values.get(option_bag.path, {}).get(option_bag.index, default_value)
        if owner == owners.default or \
                ('frozen' in option_bag.properties and \
                 ('force_default_on_freeze' in option_bag.properties or \
                 self.check_force_to_metaconfig(option_bag))):
            # the value is a default value
            # get it
            value = self.get_default_value(option_bag)
        value, has_calculation = self.get_calculated_value(option_bag,
                                                           value,
                                                           )
        return value, has_calculation

    def get_default_value(self,
                          option_bag: OptionBag,
                          ) -> Any:
        """get default value:
        - get parents config value or
        - get calculated value or
        - get default value
        """
        moption_bag = self._get_modified_parent(option_bag)
        if moption_bag is not None:
            # retrieved value from parent config
            return moption_bag.config_bag.context.get_values().get_cached_value(moption_bag)

        # now try to get calculated value:
        value, _has_calculation = self.get_calculated_value(option_bag,
                                                            option_bag.option.impl_getdefault(),
                                                            )
        if option_bag.index is not None and isinstance(value, (list, tuple)) \
                and (not option_bag.option.impl_is_submulti() or \
                not value or isinstance(value[0], list)):
            # if index (so slave), must return good value for this index
            # for submulti, first index is a list, assume other data are list too
            if len(value) > option_bag.index:
                value = value[option_bag.index]
            else:
                # no value for this index, retrieve default multi value
                # default_multi is already a list for submulti
                value, _has_calculation = self.get_calculated_value(option_bag,
                                                                    option_bag.option.impl_getdefault_multi(),
                                                                    )
        return value

    def get_calculated_value(self,
                             option_bag,
                             value,
                             reset_cache=True,
                             ) -> Any:
        """value could be a calculation, in this case do calculation
        """
        has_calculation = False
        if isinstance(value, Calculation):
            try:
                value = value.execute(option_bag)
            except ConfigError as err:
                msg = _(f'error when calculating "{option_bag.option.impl_get_display_name()}": '
                        f'{err} : {option_bag.path}')
                raise ConfigError(msg) from err
            has_calculation = True
        elif isinstance(value, list):
            # if value is a list, do subcalculation
            for idx, val in enumerate(value):
                value[idx], _has_calculation = self.get_calculated_value(option_bag,
                                                                         val,
                                                                         reset_cache=False,
                                                                         )
                if _has_calculation:
                    has_calculation = True
        if reset_cache:
            self.reset_cache_after_calculation(option_bag,
                                               value,
                                               )
        return value, has_calculation

    #______________________________________________________________________
    def check_force_to_metaconfig(self,
                                  option_bag: OptionBag,
                                  ) -> bool:
        """Check if the value must be retrieve from parent metaconfig or not
        """
        # force_metaconfig_on_freeze is set to an option and context is a kernelconfig
        #   => to metaconfig
        # force_metaconfig_on_freeze is set *explicitly* to an option and context is a
        #   kernelmetaconfig => to sub metaconfig
        if 'force_metaconfig_on_freeze' in option_bag.properties:
            settings = option_bag.config_bag.context.get_settings()
            if option_bag.config_bag.context.impl_type == 'config':
                return True
            # it's a not a config, force to metaconfig only in *explicitly* set
            return 'force_metaconfig_on_freeze' in settings.get_stored_properties(option_bag.path,
                                                                                  option_bag.index,
                                                                                  frozenset(),
                                                                                  )
        return False

    def reset_cache_after_calculation(self,
                                      option_bag,
                                      value,
                                      ):
        """if value is modification after calculation, invalid cache
        """
        cache = option_bag.config_bag.context.get_values_cache()
        is_cache, cache_value, _ = cache.getcache(option_bag,
                                                  'values',
                                                  expiration=False,
                                                  )
        if not is_cache or cache_value == value:
            # calculation return same value as previous value,
            # so do not invalidate cache
            return
        # calculated value is a new value, so reset cache
        option_bag.config_bag.context.reset_cache(option_bag)
        # and manage force_store_value
        self._set_force_value_suffix(option_bag)

    def isempty(self,
                option_bag: OptionBag,
                value: Any,
                force_allow_empty_list: bool,
                ) -> bool:
        """convenience method to know if an option is empty
        """
        if option_bag.index is None and option_bag.option.impl_is_submulti():
            # index is not set
            isempty = True
            for val in value:
                isempty = self._isempty_multi(val, force_allow_empty_list)
                if isempty:
                    break
        elif (option_bag.index is None or \
                (option_bag.index is not None and option_bag.option.impl_is_submulti())) and \
                option_bag.option.impl_is_multi():
            # it's a single list
            isempty = self._isempty_multi(value, force_allow_empty_list)
        else:
            isempty = value is None or value == ''
        return isempty

    def _isempty_multi(self,
                       value: Any,
                       force_allow_empty_list: bool,
                       ) -> bool:
        if not isinstance(value, list):
            return False
        return (not force_allow_empty_list and value == []) or None in value or '' in value

    #______________________________________________________________________
    # set value
    def set_value(self,
                  option_bag: OptionBag,
                  value: Any,
                  ) -> None:
        """set value to option
        """
        owner = self.get_context_owner()
        setting_properties = option_bag.config_bag.properties
        ori_value = value
        if 'validator' in setting_properties:
            value, has_calculation = self.setvalue_validation(value,
                                                              option_bag,
                                                              )

        elif isinstance(value, list):
            # copy
            value = value.copy()
        self._setvalue(option_bag,
                       ori_value,
                       owner,
                       )
        validator = 'validator' in setting_properties and \
                'demoting_error_warning' not in setting_properties
        if validator and not has_calculation:
            cache = option_bag.config_bag.context.get_values_cache()
            cache.setcache(option_bag,
                           value,
                           validated=validator,
                           )
        elif 'validator' in setting_properties and has_calculation:
            cache = option_bag.config_bag.context.get_values_cache()
            cache.delcache(option_bag.path)
        if 'force_store_value' in setting_properties and option_bag.option.impl_is_leader():
            leader = option_bag.option.impl_get_leadership()
            leader.follower_force_store_value(value,
                                              option_bag.config_bag,
                                              owners.forced,
                                              )

    def setvalue_validation(self,
                            value,
                            option_bag,
                            ):
        """validate value before set value
        """
        settings = option_bag.config_bag.context.get_settings()
        # First validate properties with this value
        opt = option_bag.option
        settings.validate_frozen(option_bag)
        val, has_calculation = self.get_calculated_value(option_bag,
                                                         value,
                                                         False,
                                                         )
        settings.validate_mandatory(val,
                                    option_bag,
                                    )
        # Value must be valid for option
        opt.impl_validate(val,
                          option_bag,
                          check_error=True,
                          )
        if 'warnings' in option_bag.config_bag.properties:
            # No error found so emit warnings
            opt.impl_validate(val,
                              option_bag,
                              check_error=False,
                              )
        return val, has_calculation

    def _setvalue(self,
                  option_bag: OptionBag,
                  value: Any,
                  owner: str,
                  ) -> None:
        option_bag.config_bag.context.reset_cache(option_bag)
        self.set_storage_value(option_bag.path,
                               option_bag.index,
                               value,
                               owner,
                               )
        self._set_force_value_suffix(option_bag)

    def set_storage_value(self,
                          path,
                          index,
                          value,
                          owner,
                          ):
        """set a value
        """
        self._values.setdefault(path, {})[index] = [value, owner]

    def _set_force_value_suffix(self, option_bag: OptionBag) -> None:
        """ force store value for an option for suffixes
        """
        # pylint: disable=too-many-locals
        if 'force_store_value' not in option_bag.config_bag.properties:
            return

        for woption in option_bag.option._get_suffixes_dependencies():  # pylint: disable=protected-access
            # options from dependencies are weakref
            option = woption()
            force_store_options = []
            for coption in option.get_children_recursively(None,
                                                           None,
                                                           option_bag.config_bag,
                                                           ):
                if 'force_store_value' in coption.impl_getproperties():
                    force_store_options.append(coption)
            if not force_store_options:
                continue
            for coption in force_store_options:
                if coption.impl_is_follower():
                    leader = coption.impl_get_leadership().get_leader()
                    loption_bag = OptionBag(leader,
                                            None,
                                            option_bag.config_bag,
                                            properties=frozenset(),
                                            )
                    indexes = range(len(self.get_value(loption_bag)[0]))
                else:
                    indexes = [None]
                for index in indexes:
                    for coption_bag in option.get_sub_children(coption,
                                                               option_bag.config_bag,
                                                               index=index,
                                                               properties=frozenset(),
                                                               ):
                        default_value = [self.get_value(coption_bag)[0], owners.forced]
                        self._values.setdefault(coption_bag.path, {})[index] = default_value

    def _get_modified_parent(self,
                             option_bag: OptionBag,
                             ) -> Optional[OptionBag]:
        """ Search in differents parents a Config with a modified value
        If not found, return None
        For follower option, return the Config where leader is modified
        """
        def build_option_bag(option_bag, parent):
            doption_bag = option_bag.copy()
            config_bag = option_bag.config_bag.copy()
            config_bag.context = parent
            config_bag.unrestraint()
            doption_bag.config_bag = config_bag
            return doption_bag

        for parent in option_bag.config_bag.context.get_parents():
            doption_bag = build_option_bag(option_bag, parent)
            if 'force_metaconfig_on_freeze' in option_bag.properties:
                # remove force_metaconfig_on_freeze only if option in metaconfig
                # hasn't force_metaconfig_on_freeze properties
                ori_properties = doption_bag.properties
                settings = doption_bag.config_bag.context.get_settings()
                doption_bag.properties = settings.getproperties(doption_bag)
                if not self.check_force_to_metaconfig(doption_bag):
                    doption_bag.properties = ori_properties - {'force_metaconfig_on_freeze'}
                else:
                    doption_bag.properties = ori_properties
            parent_owner = parent.get_values().getowner(doption_bag,
                                                        only_default=True,
                                                        )
            if parent_owner != owners.default:
                return doption_bag

        return None


    #______________________________________________________________________
    # owner

    def is_default_owner(self,
                         option_bag: OptionBag,
                         validate_meta: bool=True,
                         ) -> bool:
        """is default owner for an option
        """
        return self.getowner(option_bag,
                             validate_meta=validate_meta,
                             only_default=True,
                             ) == owners.default

    def hasvalue(self,
                 path,
                 index=None,
                 ):
        """if path has a value
        return: boolean
        """
        has_path = path in self._values
        if index is None:
            return has_path
        if has_path:
            return index in self._values[path]
        return False

    def getowner(self,
                 option_bag,
                 validate_meta=True,
                 only_default=False,
                 ):
        """
        retrieves the option's owner

        :param opt: the `option.Option` object
        :param force_permissive: behaves as if the permissive property
                                 was present
        :returns: a `setting.owners.Owner` object
        """
        context = option_bag.config_bag.context
        opt = option_bag.option
        if opt.impl_is_symlinkoption():
            option_bag.ori_option = opt
            opt = opt.impl_getopt()
            option_bag.option = opt
            option_bag.path = opt.impl_getpath()
        settings = context.get_settings()
        settings.validate_properties(option_bag)
        if 'frozen' in option_bag.properties and \
                'force_default_on_freeze' in option_bag.properties:
            return owners.default
        if only_default:
            if self.hasvalue(option_bag.path,
                             index=option_bag.index,
                             ):
                owner = 'not_default'
            else:
                owner = owners.default
        else:
            owner = self._values.get(option_bag.path, {}).get(option_bag.index,
                                                              [undefined, owners.default],
                                                              )[1]
        if validate_meta is not False and (owner is owners.default or
                                           'frozen' in option_bag.properties and
                                           'force_metaconfig_on_freeze' in option_bag.properties):
            moption_bag = self._get_modified_parent(option_bag)
            if moption_bag is not None:
                values = moption_bag.config_bag.context.get_values()
                owner = values.getowner(moption_bag,
                                        only_default=only_default,
                                        )
            elif 'force_metaconfig_on_freeze' in option_bag.properties:
                return owners.default
        return owner

    def set_owner(self,
                  option_bag,
                  owner,
                  ):
        """
        sets a owner to an option

        :param option_bag: the `OptionBag` object
        :param owner: a valid owner, that is a `setting.owners.Owner` object
        """
        if owner in forbidden_owners:
            raise ValueError(_('set owner "{0}" is forbidden').format(str(owner)))

        if not self.hasvalue(option_bag.path,
                             index=option_bag.index,
                             ):
            raise ConfigError(_(f'no value for {option_bag.path} cannot change owner to {owner}'))
        option_bag.config_bag.context.get_settings().validate_frozen(option_bag)
        self._values[option_bag.path][option_bag.index][1] = owner
    #______________________________________________________________________
    # reset

    def reset(self, option_bag: OptionBag) -> None:
        """reset value for an option
        """
        context = option_bag.config_bag.context
        hasvalue = self.hasvalue(option_bag.path)
        setting_properties = option_bag.config_bag.properties

        if hasvalue and 'validator' in option_bag.config_bag.properties:
            fake_context = context.gen_fake_values()
            config_bag = option_bag.config_bag.copy()
            config_bag.remove_validation()
            config_bag.context = fake_context
            soption_bag = option_bag.copy()
            soption_bag.config_bag = config_bag
            fake_value = fake_context.get_values()
            fake_value.reset(soption_bag)
            soption_bag.config_bag.properties = option_bag.config_bag.properties
            value = fake_value.get_default_value(soption_bag)
            fake_value.setvalue_validation(value,
                                           soption_bag,
                                           )
        opt = option_bag.option
        if opt.impl_is_leader():
            opt.impl_get_leadership().reset(option_bag.config_bag)
        if hasvalue:
            if 'force_store_value' in option_bag.config_bag.properties and \
                    'force_store_value' in option_bag.properties:
                value = self.get_default_value(option_bag)

                self._setvalue(option_bag,
                               value,
                               owners.forced,
                               )
            else:
                # for leader only
                value = None
                if option_bag.path in self._values:
                    del self._values[option_bag.path]
            context.reset_cache(option_bag)
            if 'force_store_value' in setting_properties and option_bag.option.impl_is_leader():
                if value is None:
                    value = self.get_default_value(option_bag)
                leader = option_bag.option.impl_get_leadership()
                leader.follower_force_store_value(value,
                                                  option_bag.config_bag,
                                                  owners.forced,
                                                  )
    #______________________________________________________________________
    # Follower

    def get_max_length(self, path: str) -> int:
        """get max index for a follower and determine the length of the follower
        """
        values = self._values.get(path, {})
        if values:
            return max(values) + 1
        return 0

    def reset_follower(self,
                       option_bag: OptionBag,
                       ) -> None:
        """reset value for a follower
        """
        if not self.hasvalue(option_bag.path,
                             index=option_bag.index,
                             ):
            return
        context = option_bag.config_bag.context
        setting_properties = option_bag.config_bag.properties
        if 'validator' in setting_properties:
            fake_context = context.gen_fake_values()
            fake_value = fake_context.get_values()
            config_bag = option_bag.config_bag.copy()
            config_bag.remove_validation()
            config_bag.context = fake_context
            soption_bag = option_bag.copy()
            soption_bag.config_bag = config_bag
            fake_value.reset_follower(soption_bag)
            value = fake_value.get_default_value(soption_bag)
            fake_value.setvalue_validation(value,
                                           soption_bag)
        if 'force_store_value' in setting_properties and \
                'force_store_value' in option_bag.properties:
            value = self.get_default_value(option_bag)

            self._setvalue(option_bag,
                           value,
                           owners.forced,
                           )
        else:
            self.resetvalue_index(option_bag)
        context.reset_cache(option_bag)

    def resetvalue_index(self, option_bag: OptionBag) -> None:
        """reset a value for a follower at an index
        """
        if option_bag.path in self._values and option_bag.index in self._values[option_bag.path]:
            del self._values[option_bag.path][option_bag.index]

    def reduce_index(self, option_bag: OptionBag) -> None:
        """reduce follower's value from a specified index
        """
        self.resetvalue_index(option_bag)
        for index in range(option_bag.index + 1, self.get_max_length(option_bag.path)):
            if self.hasvalue(option_bag.path,
                             index=index,
                             ):
                self._values[option_bag.path][index - 1] = self._values[option_bag.path].pop(index)

    def reset_leadership(self,
                         option_bag: OptionBag,
                         leadership_option_bag: OptionBag,
                         index: int,
                         ) -> None:
        """reset leadershop from an index
        """
        current_value = self.get_cached_value(option_bag)
        length = len(current_value)
        if index >= length:
            raise IndexError(_('index {index} is greater than the length {length} '
                               'for option "{option_bag.option.impl_get_display_name()}"'))
        current_value.pop(index)
        leadership_option_bag.option.pop(index,
                                         option_bag.config_bag,
                                         )
        self.set_value(option_bag,
                       current_value,
                       )

    #______________________________________________________________________
    # information

    def set_information(self,
                        option_bag,
                        key,
                        value,
                        ):
        """updates the information's attribute

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        if option_bag is None:
            path = None
        else:
            path = option_bag.path
        self._informations.setdefault(path, {})[key] = value
        if path is None:
            return
        for key, options in option_bag.option.get_dependencies_information().items():
            for option in options:
                cache_option_bag = OptionBag(option,
                                             None,
                                             option_bag.config_bag,
                                             properties=None,
                                             )
                cache_option_bag.config_bag.context.reset_cache(cache_option_bag)

    def get_information(self,
                        option_bag,
                        name,
                        default,
                        ):
        """retrieves one information's item

        :param name: the item string (ex: "help")
        """
        if option_bag is None:
            path = None
        else:
            path = option_bag.path
        try:
            return self._informations[path][name]
        except KeyError as err:
            if option_bag:
                return option_bag.option.impl_get_information(name, default)
            if default is not undefined:
                return default
            raise ValueError(_("information's item not found: {0}").format(name)) from err

    def del_information(self,
                        key: Any,
                        raises: bool=True,
                        path: str=None,
                        ):
        """delete information for a specified key
        """
        if path in self._informations and key in self._informations[path]:
            del self._informations[path][key]
        elif raises:
            raise ValueError(_(f"information's item not found \"{key}\""))

    def list_information(self,
                         path: str=None,
                         ) -> List[str]:
        """list all informations keys for a specified path
        """
        return list(self._informations.get(path, {}).keys())

    #____________________________________________________________
    # default owner methods
    def set_context_owner(self, owner: str) -> None:
        """set the context owner
        """
        if owner in forbidden_owners:
            raise ValueError(_('set owner "{0}" is forbidden').format(str(owner)))
        self._values[None][None][1] = owner

    def get_context_owner(self) -> str:
        """get the context owner
        """
        return self._values[None][None][1]



# -*- coding: utf-8 -*-
# Copyright (C) 2018-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""SynDynOption internal option, it's an instanciate synoption
"""
from typing import Any
from .baseoption import BaseOption


class SynDynOption:
    """SynDynOption is an Option include un DynOptionDescription with specified prefix
    """
    __slots__ = ('rootpath',
                 'opt',
                 'suffix',
                 'dyn_parent',
                 '__weakref__')

    def __init__(self,
                 opt: BaseOption,
                 rootpath: str,
                 suffix: str,
                 dyn_parent,
                 ) -> None:
        self.opt = opt
        self.rootpath = rootpath
        self.suffix = suffix
        self.dyn_parent = dyn_parent

    def __getattr__(self,
                    name: str) -> Any:
        return getattr(self.opt,
                       name,
                       )

    def impl_getname(self) -> str:
        """get option name
        """
        return self.opt.impl_getname()

    def impl_get_display_name(self) -> str:
        """get option display name
        """
        return self.opt.impl_get_display_name(self)

    def impl_getsuffix(self) -> str:
        """get suffix
        """
        return self.suffix

    def impl_getpath(self) -> str:
        """get path
        """
        path = self.impl_getname()
        if self.rootpath:
            path = f'{self.rootpath}.{path}'
        return path

    def impl_is_dynsymlinkoption(self) -> bool:
        """it's a dynsymlinkoption
        """
        return True

    def impl_get_leadership(self):  # pylint: disable=inconsistent-return-statements
        """is it a leadership?
        """
        leadership = self.opt.impl_get_leadership()
        if leadership:
            rootpath = self.rootpath.rsplit('.', 1)[0]
            return leadership.to_dynoption(rootpath,
                                           self.suffix,
                                           self.dyn_parent,
                                           )



# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Configuration management library written in python
"""
from .function import calc_value, calc_value_property_help, valid_ip_netmask, \
                      valid_network_netmask, valid_in_network, valid_broadcast, \
                      valid_not_equal, function_waiting_for_dict
from .autolib import Calculation, Params, ParamOption, ParamDynOption, ParamSelfOption, \
                     ParamValue, ParamIndex, ParamSuffix, ParamInformation, ParamSelfInformation
from .option import *
from .error import ConfigError
from .api import Config, MetaConfig, GroupConfig, MixConfig
from .option import __all__ as all_options
from .setting import owners, groups, undefined


allfuncs = ['Calculation',
            'Params',
            'ParamOption',
            'ParamDynOption',
            'ParamSelfOption',
            'ParamValue',
            'ParamIndex',
            'ParamSuffix',
            'ParamInformation',
            'ParamSelfInformation',
            'MetaConfig',
            'MixConfig',
            'GroupConfig',
            'Config',
            'ConfigError',
            'undefined',
            'owners',
            'groups',
            'calc_value',
            'calc_value_property_help',
            'valid_ip_netmask',
            'valid_network_netmask',
            'valid_in_network',
            'valid_broadcast',
            'function_waiting_for_dict',
            ]
allfuncs.extend(all_options)
del(all_options)
__all__ = tuple(allfuncs)
del(allfuncs)
__version__ = "4.0"



# -*- coding: utf-8 -*-
"logger for tiramisu"
# Copyright (C) 2019-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from logging import getLogger, DEBUG, StreamHandler, Formatter
import os


log = getLogger('tiramisu')
if os.environ.get('TIRAMISU_DEBUG') == 'True':  # pragma: no cover
    log.setLevel(DEBUG)
    handler = StreamHandler()
    handler.setLevel(DEBUG)
    formatter = Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    log.addHandler(handler)



# -*- coding: utf-8 -*-
"cache used by storage"
# Copyright (C) 2013-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from time import time


class Cache:
    """cache object
    """
    __slots__ = ('_cache',)

    def __init__(self):
        self._cache = {}

    def _get_path_index(self, option_bag):
        if option_bag is None:
            path = None
            index = None
        else:
            path = option_bag.path
            index = option_bag.index
        return path, index

    def getcache(self,
                 option_bag,
                 type_,
                 expiration=True,
                 ):
        """get the cache value fot a specified path
        """
        no_cache = False, None, False
        path, index = self._get_path_index(option_bag)
        if path not in self._cache or index not in self._cache[path]:
            return no_cache
        value, timestamp, validated = self._cache[path][index]
        if type_ == 'context_props':
            # cached value is settings properties so value is props
            props = value
            self_props = {}
        else:
            props = option_bag.config_bag.properties
            if type_ == 'self_props':
                # cached value is self_props
                self_props = value
            else:
                self_props = option_bag.properties
        if 'cache' in props or \
                'cache' in self_props:
            if expiration and timestamp and \
                    ('expire' in props or \
                     'expire' in self_props):
                ntime = int(time())
                if timestamp + option_bag.config_bag.expiration_time >= ntime:
                    return True, value, validated
            else:
                return True, value, validated
        return no_cache

    def setcache(self,
                 option_bag,
                 val,
                 type_='values',
                 validated=True,
                 ):
        """add val in cache for a specified path
        if follower, add index
        """
        if type_ == 'values':
            if 'cache' not in option_bag.config_bag.properties and \
                    'cache' not in option_bag.properties:
                return
        elif (option_bag is None or 'cache' not in option_bag.config_bag.properties) and \
                'cache' not in val:
            return
        path, index = self._get_path_index(option_bag)
        self._cache.setdefault(path, {})[index] = (val, int(time()), validated)

    def delcache(self, path):
        """reset cache a a specified path
        """
        if path in  self._cache:
            del self._cache[path]

    def get_cached(self):
        """get cache values
        """
        return self._cache

    def reset_all_cache(self):
        """reset all cache values
        """
        self._cache.clear()



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""SynDynOptionDescription and SynDynLeadership internal option
it's an instanciate synoptiondescription
"""
from typing import Optional, Iterator, Any, List


from ..i18n import _
from ..setting import ConfigBag, undefined
from .baseoption import BaseOption
from .syndynoption import SynDynOption


class SubDynOptionDescription:
    __slots__ = ('rootpath',
                 'opt',
                 'dyn_parent',
                 '__weakref__',
                 )

    def __init__(self,
                 opt: BaseOption,
                 rootpath: str,
                 dyn_parent,
                 ) -> None:
        self.opt = opt
        self.rootpath = rootpath
        self.dyn_parent = dyn_parent

    def impl_getpath(self) -> str:
        """get path
        """
        path = self.opt.impl_getname()
        if self.rootpath:
            path = f'{self.rootpath}.{path}'
        return path

    def get_sub_children(self,
                         option,
                         config_bag,
                         *,
                         index=None,
                         properties=undefined,
                         ):
        return self.opt.get_sub_children(option,
                                         config_bag,
                                         index=index,
                                         properties=properties,
                                         dynoption=self,
                                         )

    def getsubdyn(self):
        return self.opt.getsubdyn()

    def impl_is_optiondescription(self):
        return True

    def impl_is_dynsymlinkoption(self):
        return True

    def impl_is_sub_dyn_optiondescription(self):
        return True

    def impl_get_display_name(self) -> str:
        return self.opt.impl_get_display_name(self)

    def impl_is_dynoptiondescription(self) -> bool:
        return True

    def to_dynoption(self,
                     rootpath: str,
                     suffix: str,
                     ori_dyn,
                     ):
        return self.opt.to_dynoption(rootpath, suffix, ori_dyn)


class SynDynOptionDescription:
    """SynDynOptionDescription internal option, it's an instanciate synoptiondescription
    """
    __slots__ = ('opt',
                 'rootpath',
                 '_suffix',
                 'ori_dyn')

    def __init__(self,
                 opt: BaseOption,
                 rootpath: str,
                 suffix: str,
                 ori_dyn) -> None:
        self.opt = opt
        self.rootpath = rootpath
        self._suffix = suffix
        # For a Leadership inside a DynOptionDescription
        self.ori_dyn = ori_dyn

    def __getattr__(self,
                    name: str,
                    ) -> Any:
        # if not in SynDynOptionDescription, get value in self.opt
        return getattr(self.opt,
                       name,
                       )

    def impl_getname(self) -> str:
        """get name
        """
        if self.opt.impl_is_dynoptiondescription():
            return self.opt.impl_getname(self._suffix)
        return self.opt.impl_getname()

    def impl_get_display_name(self) -> str:
        """get display name
        """
        return self.opt.impl_get_display_name(self)

    def get_children(self,
                     config_bag: ConfigBag,
                     dyn: bool=True,
                     ):
        # pylint: disable=unused-argument
        """get children
        """
        yield from self.opt.get_children(config_bag,
                                         dynoption=self,
                                         option_suffix=self._suffix,
                                         )

    def get_child(self,
                  name: str,
                  config_bag: ConfigBag,
                  subpath: str,
                  allow_dynoption: bool=False,
                  ):
        """get children
        """
        return self.opt.get_child(name,
                                  config_bag,
                                  subpath,
                                  dynoption=self,
                                  option_suffix=self._suffix,
                                  allow_dynoption=allow_dynoption,
                                  )

    def get_sub_children(self,
                         option,
                         config_bag,
                         *,
                         index=None,
                         properties=undefined,
                         ):
        return self.opt.get_sub_children(option,
                                         config_bag,
                                         index=index,
                                         properties=properties,
                                         dynoption=self,
                                         )

    def impl_is_dynsymlinkoption(self) -> bool:
        """it's a dynsymlinkoption
        """
        return True

    def get_children_recursively(self,
                                 bytype: Optional[BaseOption],
                                 byname: Optional[str],
                                 config_bag: ConfigBag,
                                 self_opt: BaseOption=None,
                                 ) -> BaseOption:
        # pylint: disable=unused-argument
        """get children recursively
        """
        for option in self.opt.get_children_recursively(bytype,
                                                        byname,
                                                        config_bag,
                                                        self,
                                                        ):
            yield option

    def impl_getpath(self) -> str:
        """get path
        """
        path = self.impl_getname()
        if self.rootpath:
            path = f'{self.rootpath}.{path}'
        return path

    def impl_getsuffix(self) -> str:
        """get suffix
        """
        return self._suffix


class SynDynLeadership(SynDynOptionDescription):
    """SynDynLeadership internal option, it's an instanciate synoptiondescription
    """
    def get_leader(self) -> SynDynOption:
        """get the leader
        """
        return self.opt.get_leader().to_dynoption(self.impl_getpath(),
                                                  self._suffix,
                                                  self.ori_dyn,
                                                  )

    def get_followers(self) -> Iterator[SynDynOption]:
        """get followers
        """
        subpath = self.impl_getpath()
        for follower in self.opt.get_followers():
            yield follower.to_dynoption(subpath,
                                        self._suffix,
                                        self.ori_dyn,
                                        )

    def reset_cache(self,
                    path: str,
                    config_bag: 'ConfigBag',
                    resetted_opts: List[str],
                    ) -> None:
        """reset cache
        """
        leader = self.get_leader()
        followers = self.get_followers()
        self._reset_cache(path,
                          leader,
                          followers,
                          config_bag,
                          resetted_opts,
                          )

    def pop(self,
            *args,
            **kwargs,
            ) -> None:
        """pop value for a follower
        """
        self.opt.pop(*args,
                     followers=self.get_followers(),
                     **kwargs,
                     )

    def follower_force_store_value(self,
                                   value,
                                   config_bag,
                                   owner,
                                   ) -> None:
        """force store value for a follower
        """
        self.opt.follower_force_store_value(value,
                                            config_bag,
                                            owner,
                                            dyn=self,
                                            )

    def impl_getsuffix(self) -> str:
        """get suffix
        """
        return self._suffix



# -*- coding: utf-8 -*-
# Copyright (C) 2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""PermissionsOption
"""
import re

from ..i18n import _
from .intoption import IntOption


class PermissionsOption(IntOption):
    """Unix file permissions
    Valid the representing Unix permissions is an octal (base-8) notation.
    This notation consists of at least three digits (owner, group, and others).
    If a fourth digit is present to the setuid bit, the setgid bit and the sticky bit attributes.
    This option is an integer value.
    """
    __slots__ = tuple()
    perm_re = re.compile(r"^[0-7]{3,4}$")
    _type = _('unix file permissions')

    def __init__(self,
                 *args,
                 **kwargs,
                 ) -> None:
        #do not display intoption attributs
        super().__init__(*args,
                         **kwargs)

    def validate(self,
                 value: str) -> None:
        super().validate(value)
        if not self.perm_re.search(str(value)):
            raise ValueError(_('only 3 or 4 octal digits are allowed'))

    def second_level_validation(self,
                                value: str,
                                warnings_only: bool) -> None:
        old_digit = 7
        str_value = str(value)
        if len(str_value) == 4:
            str_value = str_value[1:]
        for idx, digit in enumerate(str_value):
            new_digit = int(digit)
            if old_digit < new_digit:
                if idx == 1:
                    old = _('user')
                    new = _('group')
                else:
                    old = _('group')
                    new = _('other')
                raise ValueError(_(f'{new} has more right than {old}'))
            old_digit = new_digit
        if str_value == '777':
            raise ValueError(_('too weak'))
